/******************************************************************* HEADER ***/
/**                                                                           */
/** @details   This library provides functions for button debouncing on AVR   */
/**            input pins. The algorithm is developed and published by Peter  */
/**            Dannegger.                                                     */
/**                                                                           */
/** @author    Michael Neumair (e1128631@student.tuwien.ac.at)                */
/** @version   1.0                                                            */
/** @date      16.08.2015 - first implementation                              */
/**                                                                           */
/*  @toolchain avr-gcc (GCC) 4.8.1                                            */
/**                                                                           */
/**                                                                           */
/** @copyright GNU General Public License (GPL)\n                             */
/** ------------------------------------------------------------------------\n*/
/** This program is free software: you can redistribute it and/or modify      */
/** it under the terms of the GNU General Public License as published by      */
/** the Free Software Foundation, either version 3 of the License, or         */
/** (at your option) any later version.\n\n                                   */
/**                                                                           */
/** This program is distributed in the hope that it will be useful,           */
/** but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/** GNU General Public License for more details.\n\n                          */
/**                                                                           */
/** You should have received a copy of the GNU General Public License         */
/** along with this program.  If not, see <http://www.gnu.org/licenses/>.\n\n */
/**                                                                           */
/**  @file button.c ***********************************************************/

/***************************************************************** includes ***/
#include "button.h"

/****************************************************************** defines ***/
/*
 * Controller specific defines
 */
#if defined(__AVR_ATmega644__) || defined (__AVR_ATmega644A__)
	/* Timer0 register */
    #define TIM_TCCRA      TCCR0A
	#define TIM_TCCRB      TCCR0B
	#define TIM_OCR        OCR0A
	#define TIM_TIMSK      TIMSK0

	/* Timer0 masks to initialize timer */
	#define CTC_MODE       (1<<WGM01)
	#define PRESC_1        (1<<CS00)
	#define PRESC_8        (1<<CS01)
	#define PRESC_64       (1<<CS01) | (1<<CS00)
	#define PRESC_256      (1<<CS02)
	#define PRESC_1024     (1<<CS02) | (1<<CS00)

	/* Timer0 interrupt setting */
	#define TIM_COMP_vect  TIMER0_COMPA_vect
	#define TIM_OCIE       OCIE0A

	#warning The library button.h is not tested with ATmega644.

#elif defined(__AVR_ATmega8__) || defined(__AVR_ATmega16__) || defined(__AVR_ATmega32__)
	/* Timer0 register */
    #define TIM_TCCRA      TCCR0
	#define TIM_TCCRB      TCCR0
	#define TIM_OCR        OCR0
	#define TIM_TIMSK      TIMSK

	/* Timer0 masks to initialize timer */
	#define CTC_MODE       (1<<WGM01)
	#define PRESC_1        (1<<CS00)
	#define PRESC_8        (1<<CS01)
	#define PRESC_64       (1<<CS01) | (1<<CS00)
	#define PRESC_256      (1<<CS02)
	#define PRESC_1024     (1<<CS02) | (1<<CS00)

	/* Timer0 interrupt setting */
	#define TIM_COMP_vect  TIMER0_COMP_vect
	#define TIM_OCIE       OCIE0

#elif defined(__AVR_ATtiny2313__)
	/* Timer0 register */
    #define TIM_TCCRA      TCCR0A
	#define TIM_TCCRB      TCCR0B
	#define TIM_OCR        OCR0A
	#define TIM_TIMSK      TIMSK

	/* Timer0 masks to initialize timer */
	#define CTC_MODE       (1<<WGM01)
	#define PRESC_1        (1<<CS00)
	#define PRESC_8        (1<<CS01)
	#define PRESC_64       (1<<CS01) | (1<<CS00)
	#define PRESC_256      (1<<CS02)
	#define PRESC_1024     (1<<CS02) | (1<<CS00)

	/* Timer0 interrupt setting */
	#define TIM_COMP_vect  TIMER0_COMPA_vect
	#define TIM_OCIE       OCIE0A

	#warning The library button.h is not tested with ATtiny2313.

#elif defined(__AVR_ATtiny13__)
	/* Timer0 register */
    #define TIM_TCCRA      TCCR0A
	#define TIM_TCCRB      TCCR0B
	#define TIM_OCR        OCR0A
	#define TIM_TIMSK      TIMSK0

	/* Timer0 masks to initialize timer */
	#define CTC_MODE       (1<<WGM01)
	#define PRESC_1        (1<<CS00)
	#define PRESC_8        (1<<CS01)
	#define PRESC_64       (1<<CS01) | (1<<CS00)
	#define PRESC_256      (1<<CS02)
	#define PRESC_1024     (1<<CS02) | (1<<CS00)

	/* Timer0 interrupt setting */
	#define TIM_COMP_vect  TIM0_COMPA_vect
	#define TIM_OCIE       OCIE0A
#elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
	/* Timer0 register */
    #define TIM_TCCRA      TCCR0A
	#define TIM_TCCRB      TCCR0B
	#define TIM_OCR        OCR0A
	#define TIM_TIMSK      TIMSK

	/* Timer0 masks to initialize timer */
	#define CTC_MODE       (1<<WGM01)
	#define PRESC_1        (1<<CS00)
	#define PRESC_8        (1<<CS01)
	#define PRESC_64       (1<<CS01) | (1<<CS00)
	#define PRESC_256      (1<<CS02)
	#define PRESC_1024     (1<<CS02) | (1<<CS00)

	/* Timer0 interrupt setting */
	#define TIM_COMP_vect  TIM0_COMPA_vect
	#define TIM_OCIE       OCIE0A

#else
	#warning The given device has no implementation in the library button.h.
#endif

#ifndef F_CPU
	#warning F_CPU makro variable is not defined! Needed in library button.h.
#endif

/*
 * Determine prescaler settings for the timer.
 */
#if (F_CPU / (1024 * 256) < F_DEBOUNCE) && !(F_CPU / (256 * 256) < F_DEBOUNCE)
	#define TIM_PRESC PRESC_1024
	#define TIM_PRESCVAL 1024UL
#elif (F_CPU / (256 * 256) < F_DEBOUNCE) && !(F_CPU / (64 * 256) < F_DEBOUNCE)
	#define TIM_PRESC PRESC_256
	#define TIM_PRESCVAL 256UL
#elif (F_CPU / (64 * 256) < F_DEBOUNCE) && !(F_CPU / (8 * 256) < F_DEBOUNCE)
	#define TIM_PRESC PRESC_64
	#define TIM_PRESCVAL 64UL
#elif (F_CPU / (8 * 256) < F_DEBOUNCE) && !(F_CPU / (1 * 256) < F_DEBOUNCE)
	#define TIM_PRESC PRESC_8
	#define TIM_PRESCVAL 8UL
#elif (F_CPU / (1 * 256) < F_DEBOUNCE) && !(F_CPU < F_DEBOUNCE)
	#define TIM_PRESC PRESC_1
	#define TiM_PRESCVAL 1UL
#else
	#warning no valid prescaler found!
#endif

/*
 * Calculate timer compare value
 */
#define TIM_COMPVAL F_CPU / (TIM_PRESCVAL * F_DEBOUNCE) // => ~10ms


/****************************************************************** globals ***/
/*
 * Debounced and inverted state of the buttons (1 means button is pressed).
 */
volatile uint8_t debouncedState;

/*
 * Button-pressed-event flag register.
 */
volatile uint8_t pressedFlag;

/*
 * Button repeat if it is kept pressed
 */
volatile uint8_t repeatFlag;

/********************************************************************** isr ***/

ISR(TIM_COMP_vect)
{
	buttonDebounceISR();
}

/**************************************************************** functions ***/

void InitButton(void)
{
	#ifdef BUTTON1
		DDR_BUTTON  &= ~(1<<BUTTON1);
		PORT_BUTTON |= (1<<BUTTON1);
	#endif
	#ifdef BUTTON2
		DDR_BUTTON  &= ~(1<<BUTTON2);
		PORT_BUTTON |= (1<<BUTTON2);
	#endif
	#ifdef BUTTON3
		DDR_BUTTON  &= ~(1<<BUTTON3);
		PORT_BUTTON |= (1<<BUTTON3);
	#endif
	#ifdef BUTTON4
		DDR_BUTTON  &= ~(1<<BUTTON4);
		PORT_BUTTON |= (1<<BUTTON4);
	#endif
	#ifdef BUTTON5
		DDR_BUTTON  &= ~(1<<BUTTON5);
		PORT_BUTTON |= (1<<BUTTON5);
	#endif
	#ifdef BUTTON6
		DDR_BUTTON  &= ~(1<<BUTTON6);
		PORT_BUTTON |= (1<<BUTTON6);
	#endif
	#ifdef BUTTON7
		DDR_BUTTON  &= ~(1<<BUTTON7);
		PORT_BUTTON |= (1<<BUTTON7);
	#endif
	#ifdef BUTTON8
		DDR_BUTTON  &= ~(1<<BUTTON8);
		PORT_BUTTON |= (1<<BUTTON8);
	#endif

	TIM_TCCRA |= CTC_MODE;
	TIM_TCCRB |= TIM_PRESC;

	TIM_TIMSK |= (1<<TIM_OCIE); // Compare match interrupt enable

	TIM_OCR = (uint8_t)(TIM_COMPVAL);
}

/*
 * Check if a button has been pressed. Each pressed button is reported only
 * once.
 */
uint8_t buttonChanged(uint8_t mask)
{
	cli();
	{
		mask &= pressedFlag;
		pressedFlag ^= mask;
	}
	sei();

	return mask;
}

/*
 * Check if repeat functionality kicks in on held buttons. Simulate user
 * repeatedly pressing and releasing the button.
 */
uint8_t buttonIsHeld(uint8_t mask)
{
	cli();
	{
		mask &= repeatFlag;
		repeatFlag ^= mask;
	}
	sei();

	return mask;
}

/*
 * Check if the button is pressed right now
 */
uint8_t buttonIsPressed(uint8_t mask)
{
	mask &= debouncedState;
	return mask;
}

void buttonDebounceISR(void)
{
	static uint8_t counter_lbit;
	static uint8_t counter_hbit;
	static uint8_t repeatTime;

	uint8_t tmp;

	tmp = debouncedState ^ ~PIN_BUTTON;

	// Calculate new counter values
	counter_lbit = ~(counter_lbit & tmp);
	counter_hbit = counter_lbit ^ (counter_hbit & tmp);

	// Detect overflow
	tmp &= counter_lbit & counter_hbit;

	debouncedState ^= tmp;
	pressedFlag |= debouncedState & tmp;

	if( (debouncedState & REPEAT_BUTTONS) == 0)
	{
		repeatTime = REPEAT_START;
	}

	if( --repeatTime == 0 )
	{
		repeatTime = REPEAT_NEXT;
		repeatFlag = debouncedState & REPEAT_BUTTONS;
	}
}
