/******************************************************************* HEADER ***/
/**                                                                           */
/** @details   This library provides functions for button debouncing on AVR   */
/**            input pins. The algorithm is developed and published by Peter  */
/**            Dannegger.                                                     */
/**                                                                           */
/** @author    Michael Neumair (e1128631@student.tuwien.ac.at)                */
/** @version   1.0                                                            */
/** @date      16.08.2015 - first implementation                              */
/**                                                                           */
/*  @toolchain avr-gcc (GCC) 4.8.1                                            */
/**                                                                           */
/**                                                                           */
/** @copyright GNU General Public License (GPL)\n                             */
/** ------------------------------------------------------------------------\n*/
/** This program is free software: you can redistribute it and/or modify      */
/** it under the terms of the GNU General Public License as published by      */
/** the Free Software Foundation, either version 3 of the License, or         */
/** (at your option) any later version.\n\n                                   */
/**                                                                           */
/** This program is distributed in the hope that it will be useful,           */
/** but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/** GNU General Public License for more details.\n\n                          */
/**                                                                           */
/** You should have received a copy of the GNU General Public License         */
/** along with this program.  If not, see <http://www.gnu.org/licenses/>.\n\n */
/**                                                                           */
/**  @file button.h ***********************************************************/

#ifndef BUTTON_H_
#define BUTTON_H_

/***************************************************************** includes ***/
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>

/****************************************************************** defines ***/
/**
 * @brief Debounce frequency. The default value is set to 100 Hz which
 *        corresponds to a debounce interval of 10ms. It is recommended to keep
 *        this value.
 */
#define F_DEBOUNCE  100UL

/** @brief Button port data register. */
#define PORT_BUTTON PORTB
/** @brief Button ports data direction register. */
#define DDR_BUTTON  DDRB
/** @brief Input pins address of the buttons. */
#define PIN_BUTTON  PINB
/** @brief Pin number of the first button, second button... */
#define BUTTON1     PB0
#define BUTTON2     PB1
#define BUTTON3     PB2
#define BUTTON4     PB3
#define BUTTON5     PB4
//#define BUTTON6     PB5
//#define BUTTON7     PB6
//#define BUTTON8     PB7

/** @brief Define buttons with repeat functionality. */
#define REPEAT_BUTTONS (1<<BUTTON1 | 1<<BUTTON2)

/** @brief Time until repeat functionality starts to work. */
#define REPEAT_START   50 // * 10 ms
/** @brief Time between two events if repeat functionality is active. */
#define REPEAT_NEXT    20 // * 10 ms

/*************************************************************** prototypes ***/
/**
 * @fn  void InitButton(void);
 *
 * Perform initialization for button debouncing.
 *
 * @note Button debouncing needs TIMER0 of the AVR device. Make sure that the
 *       timer is not needed somewhere else!
 */
void InitButton(void);

/**
 * @fn uint8_t buttonChanged(uint8_t mask);
 *
 * Check if a button has been pressed. Each pressed button is reported only
 * once.
 *
 * @param  mask Mask needed buttons. The variable is 1 on bit position of pins
 *             that need to be checked and 0 otherwise.
 * @retval returns 1 if the button state has changed and 0 otherwise.
 */
uint8_t buttonChanged(uint8_t mask);

/**
 * @fn uint8_t buttonIsHeld(uint8_t mask);
 *
 * Check if repeat functionality kicks in on held buttons. Simulate user
 * repeatedly pressing and releasing the button.
 *
 * @param  mask Mask needed buttons. The variable is 1 on bit position of pins
 *             that need to be checked and 0 otherwise.
 * @retval returns 1 if the button state has changed and 0 otherwise.
 */
uint8_t buttonIsHeld(uint8_t mask);

/**
 * @fn uint8_t buttonIsPressed(uint8_t mask);
 *
 * Check if the button is pressed right now
 *
 * @param  mask Mask needed buttons. The variable is 1 on bit position of pins
 *             that need to be checked and 0 otherwise.
 * @retval returns 1 if the button is currently pressed and 0 otherwise.
 */
uint8_t buttonIsPressed(uint8_t mask);

/**
 * @fn void buttonDebounceISR(void);
 *
 * Debouncing functions. This function is called in the interrupt service
 * routine of the timer used for button debouncing. In standard configuration
 * TIMER0 is used and initialized automatically. Therefore this Function is just
 * needed if another timer should be used or if TIMER0 interrupt service routine
 * is defined outside of the library (e.g. if TIMER0 is used for other
 * functionalities as well).
 */
void buttonDebounceISR(void);

#endif /* BUTTON_H_ */
